@@ -5,7 +5,7 @@
  *                            | (__| |_| |  _ <| |___ 
  *                             \___|\___/|_| \_\_____|
  *
- * Copyright (C) 1998 - 2002, Daniel Stenberg, <daniel@haxx.se>, et al.
+ * Copyright (C) 1998 - 2003, Daniel Stenberg, <daniel@haxx.se>, et al.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution. The terms
@@ -18,7 +18,7 @@
  * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
  * KIND, either express or implied.
  *
- * $Id: transfer.c,v 1.132 2003/01/07 16:15:59 bagder Exp $
+ * $Id: transfer.c,v 1.143 2003/02/26 12:42:25 bagder Exp $
  ***************************************************************************/
 
 #include "setup.h"
@@ -108,6 +108,8 @@
 #define min(a, b)   ((a) < (b) ? (a) : (b))
 #endif
 
+#define CURL_TIMEOUT_EXPECT_100 1000 /* counting ms here */
+
 enum {
   KEEP_NONE,
   KEEP_READ,
@@ -231,25 +233,28 @@ CURLcode Curl_readwrite(struct connectdata *conn,
     if((k->keepon & KEEP_READ) &&
        (FD_ISSET(conn->sockfd, readfdp))) {
 
-      bool readdone = FALSE;
+      bool readdone = TRUE;
 
       /* This is where we loop until we have read everything there is to
          read or we get a EWOULDBLOCK */
       do {
+        int buffersize = data->set.buffer_size?
+          data->set.buffer_size:BUFSIZE -1;
 
-        /* read! */
-        result = Curl_read(conn, conn->sockfd, k->buf,
-                           data->set.buffer_size?
-                           data->set.buffer_size:BUFSIZE -1,
-                           &nread);
+        /* receive data from the network! */
+        result = Curl_read(conn, conn->sockfd, k->buf, buffersize, &nread);
 
         if(0>result)
           break; /* get out of loop */
         if(result>0)
           return result;
 
-        if ((k->bytecount == 0) && (k->writebytecount == 0))
+        if ((k->bytecount == 0) && (k->writebytecount == 0)) {
           Curl_pgrsTime(data, TIMER_STARTTRANSFER);
+          if(k->wait100_after_headers)
+            /* set time stamp to compare with when waiting for the 100 */
+            k->start100 = Curl_tvnow();
+        }
 
         didwhat |= KEEP_READ;
 
@@ -317,7 +322,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                 }
               }
 
-              break;		/* read more and try again */
+              break; /* read more and try again */
             }
 
             /* decrease the size of the remaining buffer */
@@ -382,9 +387,9 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
               if(100 == k->httpcode) {
                 /*
-                 * we have made a HTTP PUT or POST and this is 1.1-lingo
+                 * We have made a HTTP PUT or POST and this is 1.1-lingo
                  * that tells us that the server is OK with this and ready
-                 * to receive our stuff.
+                 * to receive the data.
                  * However, we'll get more headers now so we must get
                  * back into the header-parsing state!
                  */
@@ -459,7 +464,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                   FD_ZERO(&k->rkeepfd);
                 }
 
-                break;		/* exit header line loop */
+                break;          /* exit header line loop */
               }
 
               /* We continue reading headers, so reset the line-based
@@ -567,19 +572,29 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                   *start && isspace((int)*start);
                   start++);
 
-              /* count all non-space letters following */
-              for(end=start, len=0;
-                  *end && !isspace((int)*end);
-                  end++, len++);
+              end = strchr(start, '\r');
+              if(!end)
+                end = strchr(start, '\n');
 
-              /* allocate memory of a cloned copy */
-              data->info.contenttype = malloc(len + 1);
-              if (NULL == data->info.contenttype)
-                return CURLE_OUT_OF_MEMORY;
+              if(end) {
+                /* skip all trailing space letters */
+                for(; isspace((int)*end) && (end > start); end--);
 
-              /* copy the content-type string */
-              memcpy(data->info.contenttype, start, len);
-              data->info.contenttype[len] = 0; /* zero terminate */
+                /* get length of the type */
+                len = end-start+1;
+              
+                /* allocate memory of a cloned copy */
+                if(data->info.contenttype)
+                  free(data->info.contenttype);
+              
+                data->info.contenttype = malloc(len + 1);
+                if (NULL == data->info.contenttype)
+                  return CURLE_OUT_OF_MEMORY;
+
+                /* copy the content-type string */
+                memcpy(data->info.contenttype, start, len);
+                data->info.contenttype[len] = 0; /* zero terminate */
+              }
             }
             else if((k->httpversion == 10) &&
                     conn->bits.httpproxy &&
@@ -742,7 +757,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
              there might be a non-header part left in the end of the read
              buffer. */
 
-        }			/* end if header mode */
+        }                       /* end if header mode */
 
         /* This is not an 'else if' since it may be a rest from the header
            parsing, where the beginning of the buffer is headers and the end
@@ -831,7 +846,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                 failf(data, "Failed writing data");
                 return CURLE_WRITE_ERROR;
               }
-              failf(data, "Received problem in the chunky parser");
+              failf(data, "Received problem %d in the chunky parser", res);
               return CURLE_RECV_ERROR;
             }
             else if(CHUNKE_STOP == res) {
@@ -944,8 +959,27 @@ CURLcode Curl_readwrite(struct connectdata *conn,
           /* init the "upload from here" pointer */
           conn->upload_fromhere = k->uploadbuf;
 
-          if(!k->upload_done)
+          if(!k->upload_done) {
+            /* HTTP pollution, this should be written nicer to become more
+               protocol agnostic. */
+
+            if(k->wait100_after_headers &&
+               (conn->proto.http->sending == HTTPSEND_BODY)) {
+              /* If this call is to send body data, we must take some action:
+                 We have sent off the full HTTP 1.1 request, and we shall now
+                 go into the Expect: 100 state and await such a header */
+              k->wait100_after_headers = FALSE; /* headers sent */
+              k->write_after_100_header = TRUE; /* wait for the header */
+              FD_ZERO (&k->writefd);            /* clear it */
+              k->wkeepfd = k->writefd;          /* set the keeper variable */
+              k->keepon &= ~KEEP_WRITE;         /* disable writing */
+              k->start100 = Curl_tvnow();       /* timeout count starts now */
+              didwhat &= ~KEEP_WRITE;  /* we didn't write anything actually */
+              break;
+            }
+
             nread = fillbuffer(conn, BUFSIZE);
+          }
           else
             nread = 0; /* we're done uploading/reading */
 
@@ -964,6 +998,12 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
           /* convert LF to CRLF if so asked */
           if (data->set.crlf) {
+              if(data->state.scratch == NULL)
+                data->state.scratch = malloc(2*BUFSIZE);
+              if(data->state.scratch == NULL) {
+                failf (data, "Failed to alloc scratch buffer!");
+                return CURLE_OUT_OF_MEMORY;
+              }
             for(i = 0, si = 0; i < nread; i++, si++) {
               if (conn->upload_fromhere[i] == 0x0a) {
                 data->state.scratch[si++] = 0x0d;
@@ -998,7 +1038,13 @@ CURLcode Curl_readwrite(struct connectdata *conn,
                             &bytes_written);       /* actually send away */
         if(result)
           return result;
-        else if(conn->upload_present != bytes_written) {
+
+        if(data->set.verbose)
+          /* show the data before we change the pointer upload_fromhere */
+          Curl_debug(data, CURLINFO_DATA_OUT, conn->upload_fromhere,
+                     bytes_written);
+      
+        if(conn->upload_present != bytes_written) {
           /* we only wrote a part of the buffer (if anything), deal with it! */
 
           /* store the amount of bytes left in the buffer to write */
@@ -1023,11 +1069,6 @@ CURLcode Curl_readwrite(struct connectdata *conn,
           }
         }
 
-        if(data->set.verbose)
-          Curl_debug(data, CURLINFO_DATA_OUT, conn->upload_fromhere,
-                     bytes_written);
-      
-
         k->writebytecount += bytes_written;
         Curl_pgrsSetUploadCounter(data, (double)k->writebytecount);
 
@@ -1037,6 +1078,7 @@ CURLcode Curl_readwrite(struct connectdata *conn,
 
   } while(0); /* just to break out from! */
 
+  k->now = Curl_tvnow();
   if(didwhat) {
     /* Update read/write counters */
     if(conn->bytecountp)
@@ -1050,14 +1092,27 @@ CURLcode Curl_readwrite(struct connectdata *conn,
       /* This should allow some time for the header to arrive, but only a
          very short time as otherwise it'll be too much wasted times too
          often. */
-      k->write_after_100_header = FALSE;
-      FD_SET (conn->writesockfd, &k->writefd); /* write socket */
-      k->keepon |= KEEP_WRITE;
-      k->wkeepfd = k->writefd;
+
+      /* Quoting RFC2616, section "8.2.3 Use of the 100 (Continue) Status":
+         
+      Therefore, when a client sends this header field to an origin server
+      (possibly via a proxy) from which it has never seen a 100 (Continue)
+      status, the client SHOULD NOT wait for an indefinite period before
+      sending the request body.
+
+      */
+
+      int ms = Curl_tvdiff(k->now, k->start100);
+      if(ms > CURL_TIMEOUT_EXPECT_100) {
+        /* we've waited long enough, continue anyway */
+        k->write_after_100_header = FALSE;
+        FD_SET (conn->writesockfd, &k->writefd); /* write socket */
+        k->keepon |= KEEP_WRITE;
+        k->wkeepfd = k->writefd;
+      }
     }    
   }
 
-  k->now = Curl_tvnow();
   if(Curl_pgrsUpdate(conn))
     result = CURLE_ABORTED_BY_CALLBACK;
   else
@@ -1135,18 +1190,34 @@ CURLcode Curl_readwrite_init(struct connectdata *conn)
   /* we want header and/or body, if neither then don't do this! */
   if(conn->bits.getheader || !data->set.no_body) {
 
-    FD_ZERO (&k->readfd);		/* clear it */
+    FD_ZERO (&k->readfd);               /* clear it */
     if(conn->sockfd != -1) {
       FD_SET (conn->sockfd, &k->readfd); /* read socket */
       k->keepon |= KEEP_READ;
     }
 
-    FD_ZERO (&k->writefd);		/* clear it */
+    FD_ZERO (&k->writefd);              /* clear it */
     if(conn->writesockfd != -1) {
-      if (data->set.expect100header)
+      /* HTTP 1.1 magic:
+
+         Even if we require a 100-return code before uploading data, we might
+         need to write data before that since the REQUEST may not have been
+         finished sent off just yet.
+
+         Thus, we must check if the request has been sent before we set the
+         state info where we wait for the 100-return code
+      */
+      if (data->set.expect100header &&
+          (conn->proto.http->sending == HTTPSEND_BODY)) {
         /* wait with write until we either got 100-continue or a timeout */
         k->write_after_100_header = TRUE;
+        k->start100 = k->start;
+      }
       else {
+        if(data->set.expect100header)
+          /* when we've sent off the rest of the headers, we must await a
+             100-continue */
+          k->wait100_after_headers = TRUE;
         FD_SET (conn->writesockfd, &k->writefd); /* write socket */
         k->keepon |= KEEP_WRITE;
       }
@@ -1647,10 +1718,10 @@ CURLcode Curl_perform(struct SessionHandle *data)
 
 CURLcode 
 Curl_Transfer(struct connectdata *c_conn, /* connection data */
-              int sockfd,	/* socket to read from or -1 */
-              int size,		/* -1 if unknown at this point */
-              bool getheader,	/* TRUE if header parsing is wanted */
-              long *bytecountp,	/* return number of bytes read or NULL */
+              int sockfd,       /* socket to read from or -1 */
+              int size,         /* -1 if unknown at this point */
+              bool getheader,   /* TRUE if header parsing is wanted */
+              long *bytecountp, /* return number of bytes read or NULL */
               int writesockfd,  /* socket to write to, it may very well be
                                    the same we read from. -1 disables */
               long *writebytecountp /* return number of bytes written or
@@ -1672,11 +1743,3 @@ Curl_Transfer(struct connectdata *c_conn, /* connection data */
   return CURLE_OK;
 
 }
-          
-/*
- * local variables:
- * eval: (load-file "../curl-mode.el")
- * end:
- * vim600: fdm=marker
- * vim: et sw=2 ts=2 sts=2 tw=78
- */
